#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <memory.h>
#include <io.h>
#include "hexfile.h"

#ifdef _MSC_VER
#pragma warning(disable : 4996)
#endif

#define REC_MARK        record_buffer[0]
#define REC_SIZE        record_buffer[0]
#define REC_ADDR        MAKEWORD(record_buffer[2], record_buffer[1])
#define REC_TYPE        record_buffer[3]
#define REC_DATA        record_buffer[4]
#define EOL(c)          (((c) == '\r') || ((c) == '\n') || ((c) == '\0'))
#define X2I(c)          ((c) - ((c) <= '9' ? '0' : 'A' - 10))
#define MAKEXBYTE(l, h) (X2I(l) | (X2I(h) << 4))

static unsigned char device_buffer[DEVICE_MEMORY_SIZE];
static unsigned char verify_buffer[DEVICE_MEMORY_SIZE];
static unsigned char record_buffer[640];

unsigned char *get_device_memory(int erase)
{
  if (erase)
    memset(device_buffer, 0xFF, sizeof(device_buffer));
  return device_buffer;
}

unsigned char *get_verify_memory(int erase)
{
  if (erase)
    memset(verify_buffer, 0xFF, sizeof(verify_buffer));
  return verify_buffer;
}

static void check_range(unsigned char *buf, int addr, int size, int page, int *lo_addr, int *hi_addr)
{
  *lo_addr = addr;
  *hi_addr = addr + size;
  int off, found = 0;
  for (off = *lo_addr; !found && (off < *hi_addr); off += page)
  {
    for (int i = 0; i < page; ++i)
    {
      if (buf[off + i] != 0xFF)
      {
        found = 1;
        break;
      }
    }
  }
  *lo_addr = off;
  found = 0;
  for (off = *hi_addr; !found && (off > *lo_addr); off -= page)
  {
    for (int i = 1; i <= page; ++i)
    {
      if (buf[off - i] != 0xFF)
      {
        found = 1;
        break;
      }
    }
  }
  *hi_addr = off;
}

int verify_device_memory(int addr, int size)
{
  int count = 0;
  for (int i = 0; i < size; ++i)
  {
    if (device_buffer[addr + i] != verify_buffer[addr + i])
    {
      printf("%04X: 0x%02X <-> 0x%02X\n", addr + i, device_buffer[addr + i], verify_buffer[addr + i]);
      count++;
    }
  }
  return count;
}

void dump_device_memory(unsigned char reclen)
{
  for (int i = 0; i < sizeof(device_buffer); i += reclen)
  {
    int len = sizeof(device_buffer) - i < reclen ? sizeof(device_buffer) - i : reclen;
    for (int j = 0; j < len; ++j)
    {
      if (device_buffer[i + j] != 0xFF)
      {
        printf("%04X: ", i);
        for (j = 0; j < len; ++j)
          printf("%02X ", device_buffer[i + j]);
        printf("\n");
        break;
      }
    }
  }
}

int load_hex_file(const char *filename)
{
  int error = 1;
  FILE *fp;

  if (fp = fopen(filename, "rt"))
  {
    memset(record_buffer, 0x00, sizeof(record_buffer));
    error = 0;
    while (fgets((char *)record_buffer, sizeof(record_buffer) - 1, fp))
    {
      // intel hex format
      if (REC_MARK == ':')
      {
        unsigned char csum = 0;
        int rlen = 0;
        for (int i = 1;; )
        {
          unsigned char h = record_buffer[i++];
          unsigned char l = record_buffer[i++];
          if (!isxdigit(h))
          {
            if (!EOL(h))
            {
              printf("illegal chracter found. [%02X]\n", h);
              error = 3;
            }
            else if ((rlen < 5) || (rlen != REC_SIZE + 5))
            {
              printf("illegal record length [%d]\n", rlen);
              error = 3;
            }
            else if (csum)
            {
              printf("check sum missmatched. [%02X] != [%02X]\n", csum, record_buffer[rlen - 1]);
              error = 3;
            }
            break;
          }
          else if (!isxdigit(l))
          {
            printf("illegal chracter found. [%02X]\n", l);
            error = 3;
            break;
          }
          csum += record_buffer[rlen++] = MAKEXBYTE(l, h);
        }
        if (error)
          break;
        if (REC_TYPE == 1)
          break;
        if (REC_TYPE == 0)
          memcpy(device_buffer + REC_ADDR, &REC_DATA, REC_SIZE);
      }
      else
      {
        printf("unsupported file format. [%02X]\n", REC_MARK);
        error = 2;
        break;
      }
    }
    fclose(fp);
  }
  return error;
}

int load_bin_file(const char *filename)
{
  FILE *fp;

  if (fp = fopen(filename, "rb"))
  {
    size_t len = 16;
    for (size_t off = 0; len = fread(record_buffer, 1, len, fp); off += len)
    {
      for (size_t i = 0; i < len; ++i)
      {
        if (record_buffer[i] != 0xFF)
        {
          memcpy(device_buffer + off, record_buffer, len);
          break;
        }
      }
    }
    fclose(fp);
  }
  return fp ? 0 : 1;
}

static int save_hex_file0(const char *filename, unsigned char *buffer, int buflen, unsigned char reclen)
{
  FILE *fp;

  if (fp = fopen(filename, "wt"))
  {
    for (int i = 0; i < buflen; i += reclen)
    {
      int len = buflen - i < reclen ? buflen - i : reclen;
      for (int j = 0; j < len; ++j)
      {
        if (buffer[i + j] != 0xFF)
        {
          unsigned char csum = reclen + HIBYTE(i) + LOBYTE(i);
          fprintf(fp, ":%02X%04X00", reclen, i);
          for (j = 0; j < reclen; ++j)
          {
            unsigned char d = buffer[i + j];
            fprintf(fp, "%02X", d);
            csum += d;
          }
          fprintf(fp, "%02X\n", (~csum + 1) & 0xFF);
          break;
        }
      }
    }
    fprintf(fp, ":00000001FF");
    fclose(fp);
  }
  return fp ? 0 : 1;
}

int save_hex_file(const char *filename, unsigned char reclen)
{
  return save_hex_file0(filename, device_buffer, sizeof(device_buffer), reclen);
}

static int search_dsys_file(char *pattern, int pathlen)
{
  struct _finddata_t fileinfo;
  intptr_t handle;
  time_t filetime = 0;

  if ((handle = _findfirst(pattern, &fileinfo)) == (intptr_t)-1)
    return 1;
  do
  {
    if (filetime <= fileinfo.time_write)
    {
      if (filetime)
        remove(pattern);
      sprintf(pattern + pathlen, "%s", fileinfo.name);
      filetime = fileinfo.time_write;
    }
    else
    {
      char filepath[_MAX_PATH];
      sprintf(filepath, "%.*s%.*s", pathlen, pattern, (int)(sizeof(filepath) - pathlen - 1), fileinfo.name);
      remove(filepath);
    }
  }
  while (_findnext(handle, &fileinfo) == 0);
  _findclose(handle);
  return 0;
}

static int make_dsys_filename(char *filename, const char *device, unsigned char *dsysbuf)
{
  int pathlen, off = 0;
  char *path = getenv(ENV_INFO_PATH);
  if (path)
  {
    char *p;
    off = sprintf(filename, "%s", path);
#ifdef _MSC_VER
    while (p = strchr(filename, '\\'))
      *p = '/';
#endif
    if (off == 0)
      filename[off++] = '/';
    else if (filename[off - 1] != '/')
      filename[off++] = '/';
  }
  pathlen = off;
  off += sprintf(filename + off, "%s_", device);
  // chip id
  for (int i = 0x0B; i <= 0x0F; ++i)
    off += sprintf(filename + off, "%02X", dsysbuf[i]);
  sprintf(filename + off, ".dsys");
  return pathlen;
}

int save_dsys_to_file(const char *devicename, unsigned char *dsysbuf, int dsyslen, unsigned char reclen)
{
  char filename[_MAX_PATH];
  // search dsys file
  int pathlen = make_dsys_filename(filename, "*", dsysbuf);
  if (search_dsys_file(filename, pathlen) == 0)
    remove(filename);
  // save dsys file
  make_dsys_filename(filename, devicename, dsysbuf);
  return save_hex_file0(filename, dsysbuf, dsyslen, reclen);
}

int get_devicename_from_dsys(char *devicename, unsigned char *dsysbuf)
{
  char filename[_MAX_PATH];
  *devicename = 0;
  // search dsys file
  int pathlen = make_dsys_filename(filename, "*", dsysbuf);
  if (search_dsys_file(filename, pathlen))
    return 1;
  // get device name
  *strchr(filename + pathlen, '_') = 0;
  strcpy(devicename, filename + pathlen);
  return 0;
}
