#include <windows.h>
#include "libMPSSE_spi.h"

#define FTD2XX   "ftd2xx.dll"

typedef FT_STATUS(WINAPI *LPFN_FT_Read)(FT_HANDLE ftHandle, LPVOID lpBuffer, DWORD dwBytesToRead, LPDWORD lpBytesReturned);
typedef FT_STATUS(WINAPI *LPFN_FT_Write)(FT_HANDLE ftHandle, LPVOID lpBuffer, DWORD dwBytesToWrite, LPDWORD lpBytesWritten);

static HMODULE       hFTD2XX = NULL;
static LPFN_FT_Read  lpfn_FT_Read;
static LPFN_FT_Write lpfn_FT_Write;

#define LIBMPSSE "libMPSSE.dll"

typedef FT_STATUS (__cdecl *LPFN_SPI_GetNumChannels)(uint32 *numChannels);
typedef FT_STATUS (__cdecl *LPFN_SPI_GetChannelInfo)(uint32 index, FT_DEVICE_LIST_INFO_NODE *chanInfo);
typedef FT_STATUS (__cdecl *LPFN_SPI_OpenChannel)(uint32 index, FT_HANDLE *handle);
typedef FT_STATUS (__cdecl *LPFN_SPI_InitChannel)(FT_HANDLE handle, ChannelConfig *config);
typedef FT_STATUS (__cdecl *LPFN_SPI_CloseChannel)(FT_HANDLE handle);
typedef FT_STATUS (__cdecl *LPFN_SPI_Read)(FT_HANDLE handle, uint8 *buffer, uint32 sizeToTransfer, uint32 *sizeTransfered, uint32 options);
typedef FT_STATUS (__cdecl *LPFN_SPI_Write)(FT_HANDLE handle, uint8 *buffer, uint32 sizeToTransfer, uint32 *sizeTransfered, uint32 options);
typedef FT_STATUS (__cdecl *LPFN_SPI_ReadWrite)(FT_HANDLE handle, uint8 *inBuffer, uint8 *outBuffer, uint32 sizeToTransfer, uint32 *sizeTransferred, uint32 transferOptions);
typedef FT_STATUS (__cdecl *LPFN_SPI_IsBusy)(FT_HANDLE handle, bool *state);
typedef FT_STATUS (__cdecl *LPFN_SPI_ChangeCS)(FT_HANDLE handle, uint32 options);

static HMODULE                 hLibMPSSE = NULL;
static LPFN_SPI_GetNumChannels lpfn_SPI_GetNumChannels;
static LPFN_SPI_GetChannelInfo lpfn_SPI_GetChannelInfo;
static LPFN_SPI_OpenChannel    lpfn_SPI_OpenChannel;
static LPFN_SPI_InitChannel    lpfn_SPI_InitChannel;
static LPFN_SPI_CloseChannel   lpfn_SPI_CloseChannel;
static LPFN_SPI_Read           lpfn_SPI_Read;
static LPFN_SPI_Write          lpfn_SPI_Write;
static LPFN_SPI_ReadWrite      lpfn_SPI_ReadWrite;
static LPFN_SPI_IsBusy         lpfn_SPI_IsBusy;
static LPFN_SPI_ChangeCS       lpfn_SPI_ChangeCS;

#define MPSSE_CMD_SET_DATA_BITS_LOWBYTE 0x80
#define MPSSE_CMD_SET_DATA_BITS_HIWBYTE 0x82
#define MPSSE_CMD_GET_DATA_BITS_LOWBYTE 0x81
#define MPSSE_CMD_GET_DATA_BITS_HIWBYTE 0x83
#define MPSSE_CMD_SEND_IMMEDIATE			  0x87

#define TRANSFER_OPTIONS(p) ((p) & ~(SPI_TRANSFER_OPTIONS_CHIPSELECT_ENABLE | SPI_TRANSFER_OPTIONS_CHIPSELECT_DISABLE))

#define GPIO_SPI_SCK_PIN    0x01
#define GPIO_SPI_MOSI_PIN   0x02
#define GPIO_SPI_MISO_PIN   0x04
#define GPIO_SPI_MASK       (GPIO_SPI_SCK_PIN | GPIO_SPI_MOSI_PIN | GPIO_SPI_MISO_PIN)

#define GPIO_SPI_CS_PIN(p) ((1 << (((p) & SPI_CONFIG_OPTION_CS_MASK) >> 2)) << 3)

static uint32 config_options  = 0;
static uint16 gpio_spi_cs_pin = 0;
static uint16 gpio_directions = 0;
static uint16 gpio_out_values = 0;

static void Load_LibMPSSE()
{
  if (!hFTD2XX)
  {
    if (hFTD2XX = LoadLibrary(FTD2XX))
    {
      lpfn_FT_Read  = (LPFN_FT_Read )GetProcAddress(hFTD2XX, "FT_Read" );
      lpfn_FT_Write = (LPFN_FT_Write)GetProcAddress(hFTD2XX, "FT_Write");
    }
    else
    {
      lpfn_FT_Read  = NULL;
      lpfn_FT_Write = NULL;
    }
  }
  if (!hLibMPSSE)
  {
    if (hLibMPSSE = LoadLibrary(LIBMPSSE))
    {
      lpfn_SPI_GetNumChannels = (LPFN_SPI_GetNumChannels)GetProcAddress(hLibMPSSE, "SPI_GetNumChannels");
      lpfn_SPI_GetChannelInfo = (LPFN_SPI_GetChannelInfo)GetProcAddress(hLibMPSSE, "SPI_GetChannelInfo");
      lpfn_SPI_OpenChannel    = (LPFN_SPI_OpenChannel   )GetProcAddress(hLibMPSSE, "SPI_OpenChannel"   );
      lpfn_SPI_InitChannel    = (LPFN_SPI_InitChannel   )GetProcAddress(hLibMPSSE, "SPI_InitChannel"   );
      lpfn_SPI_CloseChannel   = (LPFN_SPI_CloseChannel  )GetProcAddress(hLibMPSSE, "SPI_CloseChannel"  );
      lpfn_SPI_Read           = (LPFN_SPI_Read          )GetProcAddress(hLibMPSSE, "SPI_Read"          );
      lpfn_SPI_Write          = (LPFN_SPI_Write         )GetProcAddress(hLibMPSSE, "SPI_Write"         );
      lpfn_SPI_ReadWrite      = (LPFN_SPI_ReadWrite     )GetProcAddress(hLibMPSSE, "SPI_ReadWrite"     );
      lpfn_SPI_IsBusy         = (LPFN_SPI_IsBusy        )GetProcAddress(hLibMPSSE, "SPI_IsBusy"        );
      lpfn_SPI_ChangeCS       = (LPFN_SPI_ChangeCS      )GetProcAddress(hLibMPSSE, "SPI_ChangeCS"      );
    }
    else
    {
      lpfn_SPI_GetNumChannels = NULL;
      lpfn_SPI_GetChannelInfo = NULL;
      lpfn_SPI_OpenChannel    = NULL;
      lpfn_SPI_InitChannel    = NULL;
      lpfn_SPI_CloseChannel   = NULL;
      lpfn_SPI_Read           = NULL;
      lpfn_SPI_Write          = NULL;
      lpfn_SPI_ReadWrite      = NULL;
      lpfn_SPI_IsBusy         = NULL;
      lpfn_SPI_ChangeCS       = NULL;
    }
  }
}

static void Free_LibMPSSE()
{
  if (hLibMPSSE)
  {
    FreeLibrary(hLibMPSSE);
    hLibMPSSE = NULL;
  }
  if (hFTD2XX)
  {
    FreeLibrary(hFTD2XX);
    hFTD2XX = NULL;
  }
}

FTDI_API void Init_libMPSSE(void)
{
  Load_LibMPSSE();
}

FTDI_API void Cleanup_libMPSSE(void)
{
  Free_LibMPSSE();
}

FTDI_API FT_STATUS SPI_GetNumChannels(uint32 *numChannels)
{
  return lpfn_SPI_GetNumChannels ? (*lpfn_SPI_GetNumChannels)(numChannels) : FT_NOT_SUPPORTED;
}

FTDI_API FT_STATUS SPI_GetChannelInfo(uint32 index, FT_DEVICE_LIST_INFO_NODE *chanInfo)
{
  return lpfn_SPI_GetChannelInfo ? (*lpfn_SPI_GetChannelInfo)(index, chanInfo) : FT_NOT_SUPPORTED;
}

FTDI_API FT_STATUS SPI_OpenChannel(uint32 index, FT_HANDLE *handle)
{
  return lpfn_SPI_OpenChannel ? (*lpfn_SPI_OpenChannel)(index, handle) : FT_NOT_SUPPORTED;
}

FTDI_API FT_STATUS SPI_InitChannel(FT_HANDLE handle, ChannelConfig *config)
{
  FT_STATUS status = FT_NOT_SUPPORTED;
  if (lpfn_SPI_InitChannel)
  {
    config_options = config->configOptions;
    uint32 pin     = config->Pin;
    uint16 cs_pin  = GPIO_SPI_CS_PIN(config_options);
    config->Pin   &= ~(GPIO_SPI_MASK | cs_pin);
    if (FT_SUCCESS(status = (*lpfn_SPI_InitChannel)(handle, config)))
    {
      gpio_directions = (((pin >> 0) & 0xFF) & ~GPIO_SPI_MASK) | (GPIO_SPI_SCK_PIN | GPIO_SPI_MOSI_PIN | cs_pin);
      gpio_out_values = (((pin >> 8) & 0xFF) & ~GPIO_SPI_MASK);
      SPI_ChangeCS(handle, config_options);
    }
  }
  return status;
}

FTDI_API FT_STATUS SPI_CloseChannel(FT_HANDLE handle)
{
  return lpfn_SPI_CloseChannel ? (*lpfn_SPI_CloseChannel)(handle) : FT_NOT_SUPPORTED;
}

FTDI_API FT_STATUS SPI_Read(FT_HANDLE handle, uint8 *buffer, uint32 sizeToTransfer, uint32 *sizeTransferred, uint32 options)
{
  FT_STATUS status = FT_NOT_SUPPORTED;
  if (lpfn_SPI_Write)
  {
    if (options & SPI_TRANSFER_OPTIONS_CHIPSELECT_ENABLE)
      SPI_ToggleCS(handle, TRUE);
    status = (*lpfn_SPI_Read)(handle, buffer, sizeToTransfer, sizeTransferred, TRANSFER_OPTIONS(options));
    if (options & SPI_TRANSFER_OPTIONS_CHIPSELECT_DISABLE)
      SPI_ToggleCS(handle, FALSE);
  }
  return status;
}

static void update_gpio_state(uint8 *buffer, uint32 sizeToTransfer, uint32 options)
{
  if (options & SPI_TRANSFER_OPTIONS_SIZE_IN_BITS)
  {
    if (buffer[sizeToTransfer >> 3] & (0x80 >> (sizeToTransfer & 7)))
      gpio_out_values |= GPIO_SPI_MOSI_PIN;
    else
      gpio_out_values &= ~GPIO_SPI_MOSI_PIN;
  }
  else
  {
    if (buffer[sizeToTransfer - 1] & 0x01)
      gpio_out_values |= GPIO_SPI_MOSI_PIN;
    else
      gpio_out_values &= ~GPIO_SPI_MOSI_PIN;
  }
}

FTDI_API FT_STATUS SPI_Write(FT_HANDLE handle, uint8 *buffer, uint32 sizeToTransfer, uint32 *sizeTransferred, uint32 options)
{
  FT_STATUS status = FT_NOT_SUPPORTED;
  if (lpfn_SPI_Write)
  {
    if (options & SPI_TRANSFER_OPTIONS_CHIPSELECT_ENABLE)
      SPI_ToggleCS(handle, TRUE);
    if (FT_SUCCESS(status = (*lpfn_SPI_Write)(handle, buffer, sizeToTransfer, sizeTransferred, TRANSFER_OPTIONS(options))))
    {
      if (sizeToTransfer && (sizeToTransfer == *sizeTransferred))
        update_gpio_state(buffer, sizeToTransfer, options);
    }
    if (options & SPI_TRANSFER_OPTIONS_CHIPSELECT_DISABLE)
      SPI_ToggleCS(handle, FALSE);
  }
  return status;
}

FTDI_API FT_STATUS SPI_ReadWrite(FT_HANDLE handle, uint8 *inBuffer, uint8 *outBuffer, uint32 sizeToTransfer, uint32 *sizeTransferred, uint32 options)
{
  FT_STATUS status = FT_NOT_SUPPORTED;
  if (lpfn_SPI_ReadWrite)
  {
    if (options & SPI_TRANSFER_OPTIONS_CHIPSELECT_ENABLE)
      SPI_ToggleCS(handle, TRUE);
    if (FT_SUCCESS(status = (*lpfn_SPI_ReadWrite)(handle, inBuffer, outBuffer, sizeToTransfer, sizeTransferred, TRANSFER_OPTIONS(options))))
    {
      if (sizeToTransfer)
        update_gpio_state(outBuffer, sizeToTransfer, options);
    }
    if (options & SPI_TRANSFER_OPTIONS_CHIPSELECT_DISABLE)
      SPI_ToggleCS(handle, FALSE);
  }
  return status;
}

FTDI_API FT_STATUS SPI_IsBusy(FT_HANDLE handle, bool *state)
{
  return lpfn_SPI_IsBusy ? (*lpfn_SPI_IsBusy)(handle, state) : FT_NOT_SUPPORTED;
}

FTDI_API FT_STATUS SPI_ChangeCS(FT_HANDLE handle, uint32 options)
{
  SPI_ToggleCS(handle, FALSE);
#if 0
  config_options = options;
#else
  config_options &=  SPI_CONFIG_OPTION_MODE_MASK;
  config_options |= ~SPI_CONFIG_OPTION_MODE_MASK & options;
#endif
  gpio_directions |= (gpio_spi_cs_pin = GPIO_SPI_CS_PIN(options));
  switch (config_options & SPI_CONFIG_OPTION_MODE_MASK)
  {
    case SPI_CONFIG_OPTION_MODE0:
    case SPI_CONFIG_OPTION_MODE1:
      gpio_out_values &= ~GPIO_SPI_SCK_PIN;
      break;
    case SPI_CONFIG_OPTION_MODE2:
    case SPI_CONFIG_OPTION_MODE3:
      gpio_out_values |= GPIO_SPI_SCK_PIN;
      break;
  }
  return SPI_ToggleCS(handle, FALSE);
}

FTDI_API FT_STATUS SPI_ToggleCS(FT_HANDLE handle, bool state)
{
  if (config_options & SPI_CONFIG_OPTION_CS_ACTIVELOW ? !state : state)
    gpio_out_values |= gpio_spi_cs_pin;
  else
    gpio_out_values &= ~gpio_spi_cs_pin;
  return FT_WriteGPIO(handle, gpio_directions, gpio_out_values);
}

FTDI_API FT_STATUS FT_WriteGPIO(FT_HANDLE handle, uint16 dir, uint16 value)
{
  FT_STATUS status = FT_NOT_SUPPORTED;
  if (lpfn_FT_Write)
  {
    value &= ~(GPIO_SPI_MASK | gpio_spi_cs_pin);
    value |=  (GPIO_SPI_MASK | gpio_spi_cs_pin) & gpio_out_values;
    dir   &= ~(GPIO_SPI_MASK | gpio_spi_cs_pin);
    dir   |=  (GPIO_SPI_MASK | gpio_spi_cs_pin) & gpio_directions;
    uint8 command[] = {MPSSE_CMD_SET_DATA_BITS_LOWBYTE, (uint8)(value & 0xFF), (uint8)(dir & 0xFF), MPSSE_CMD_SET_DATA_BITS_HIWBYTE, (uint8)(value >> 8), (uint8)(dir >> 8)};
    DWORD bytesWritten = 0;
    if (FT_SUCCESS(status = (*lpfn_FT_Write)(handle, command, sizeof(command), &bytesWritten)))
    {
      if (sizeof(command) != bytesWritten)
        status = FT_IO_ERROR;
      else
      {
        gpio_directions = dir;
        gpio_out_values = value;
      }
    }
  }
  return status;
}

FTDI_API FT_STATUS FT_ReadGPIO(FT_HANDLE handle, uint16 *value)
{
  FT_STATUS status = FT_NOT_SUPPORTED;
  if (lpfn_FT_Write && lpfn_FT_Read)
  {
    uint8 command[] = {MPSSE_CMD_GET_DATA_BITS_LOWBYTE, MPSSE_CMD_GET_DATA_BITS_HIWBYTE, MPSSE_CMD_SEND_IMMEDIATE};
    DWORD bytesTransfered = 0;
    if (FT_SUCCESS(status = (*lpfn_FT_Write)(handle, command, sizeof(command), &bytesTransfered)))
    {
      if (sizeof(command) != bytesTransfered)
        status = FT_IO_ERROR;
      else
      {
        uint8 buffer[2];
        bytesTransfered = 0;
        if (FT_SUCCESS(status = (*lpfn_FT_Read)(handle, buffer, sizeof(buffer), &bytesTransfered)))
        {
          if (sizeof(buffer) != bytesTransfered)
            status = FT_IO_ERROR;
          else
            *value = buffer[0] | ((uint16)buffer[1] << 8);
        }
      }
    }
  }
  return status;
}
