#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <ctype.h>
#include <string.h>
#include <process.h>
#include <direct.h>
#include <conio.h>
#include "hexfile.h"
#include "nrf24spi.h"
#include "variables.h"

#ifdef _MSC_VER
#pragma warning(disable : 4996)
#endif

#define LENGTHOF(p) (sizeof(p)/sizeof(p[0]))

#define MAX_COMMANDLINE_LENGTH 512
#define MAX_COMMANDLINE_ARGC   65

const char *name_24le1  = "24LE1";
const char *name_lu1p16 = "LU1P16";
const char *name_24lu1p = "24LU1P";

const char *pins_24le1 =
"  MPSSE (C232HM) <-----> * QFN24 QFN32 QFN48 \n"
"  -------------------------------------------\n"
"  DBUS0 (Orange) <SCK  > 5  P0.2  P0.5  P1.2 \n"
"  DBUS1 (Yellow) <MOSI > 4  P0.3  P0.7  P1.5 \n"
"  DBUS2 (Green ) <MISO > 7  P0.4  P1.0  P1.6 \n"
"  DBUS3 (Brown ) <CSN  > 8  P0.5  P1.1  P2.0 \n"
"  DBUS6 (White ) <PROG > 6  PROG  PROG  PROG \n"
"  DBUS7 (Blue  ) <RESET> 3  RESET RESET RESET\n"
"        (Red   ) <VDD  > 1  VDD   VDD   VDD  \n"
"        (Black ) <GND  > 9  GND   GND   GND  \n"
" *mPro-Programmer Pin's\n";

const char *pins_24lu1p =
"  MPSSE (C232HM) <-----> * QFN32 \n"
"  -------------------------------\n"
"  DBUS0 (Orange) <SCK  > 3  P0.0 \n"
"  DBUS1 (Yellow) <MOSI > 4  P0.1 \n"
"  DBUS2 (Green ) <MISO > 5  P0.2 \n"
"  DBUS3 (Brown ) <CSN  > 6  P0.3 \n"
"  DBUS6 (White ) <PROG > 2  PROG \n"
"  DBUS7 (Blue  ) <RESET>    RESET\n"
"        (Red   ) <VDD  > 1  VDD  \n"
"        (Black ) <GND  > 7  GND  \n"
" *USB-Dongle Pin's\n";

struct
{
  const char *name;
  const char *pins;
  int   page_size;
  int   dsys_addr;
  int   dsys_size;
  int   info_addr;
  int   info_size;
  int   code_addr;
  int   code_size;
  int   data_addr;
  int   data_size;
} device_spec;

int check_device_dsys(FT_HANDLE handle, char *device)
{
  FT_STATUS status;
  uint8 fsr, dsys[DSYS_SIZE];

  if (FT_SUCCESS(status = read_fsr(handle, &fsr)) && !(fsr & (FSR_RDISIP| FSR_ZERO)))
    status = read_flush(handle, 0x0000, dsys, sizeof(dsys), OPT_INFOPAGE);
  if (FT_SUCCESS(status) && !(fsr & (FSR_RDISIP | FSR_ZERO)) && (get_devicename_from_dsys(device, dsys) == 0))
    return 0;
  memset(dsys, 0xFF, sizeof(dsys));
  return get_devicename_from_dsys(device, dsys);
}

int save_device_dsys(FT_HANDLE handle)
{
  FT_STATUS status;
  uint8 fsr, dsys[DSYS_SIZE];

  if (FT_SUCCESS(status = read_fsr(handle, &fsr)) && !(fsr & (FSR_RDISIP | FSR_ZERO)))
    status = read_flush(handle, 0x0000, dsys, sizeof(dsys), OPT_INFOPAGE);
  if (!FT_SUCCESS(status) || (fsr & (FSR_RDISIP | FSR_ZERO)))
    memset(dsys, 0xFF, sizeof(dsys));
  return save_dsys_to_file(device_spec.name, dsys, sizeof(dsys), 16);
}

int change_device(FT_HANDLE handle, const char *name)
{
  int rv = 0;
  if (name)
  {
    if (_stricmp(name, name_24le1) == 0)
    {
      device_spec.name = name_24le1;
      device_spec.pins = pins_24le1;
      device_spec.page_size = 0x0200;
      device_spec.dsys_addr = 0x0000;
      device_spec.dsys_size = DSYS_SIZE;
      device_spec.info_addr = DSYS_SIZE;
      device_spec.info_size = 0x0200 - DSYS_SIZE;
      device_spec.code_addr = 0x0000;
      device_spec.code_size = 0x4000;
      device_spec.data_addr = 0x4400;
      device_spec.data_size = 0x0400;
    }
    else if (_stricmp(name, name_lu1p16) == 0)
    {
      uint8 fpcr = 0;
      read_fpcr(handle, &fpcr);
      device_spec.name = name_lu1p16;
      device_spec.pins = pins_24lu1p;
      device_spec.page_size = 0x0200;
      device_spec.dsys_addr = 0x0000;
      device_spec.dsys_size = DSYS_SIZE;
      device_spec.info_addr = DSYS_SIZE;
      device_spec.info_size = 0x0200 - DSYS_SIZE;
      device_spec.code_addr = 0x0000;
      device_spec.code_size = fpcr & 0x80 ? 0x3C00 : 0x4000;
      device_spec.data_addr = fpcr & 0x80 ? 0x7C00 : 0x0000;
      device_spec.data_size = fpcr & 0x80 ? 0x0400 : 0x0000;
    }
    else if (_stricmp(name, name_24lu1p) == 0)
    {
      uint8 fpcr = 0;
      read_fpcr(handle, &fpcr);
      device_spec.name = name_24lu1p;
      device_spec.pins = pins_24lu1p;
      device_spec.page_size = 0x0200;
      device_spec.dsys_addr = 0x0000;
      device_spec.dsys_size = DSYS_SIZE;
      device_spec.info_addr = DSYS_SIZE;
      device_spec.info_size = 0x0200 - DSYS_SIZE;
      device_spec.code_addr = 0x0000;
      device_spec.code_size = fpcr & 0x80 ? 0x7C00 : 0x8000;
      device_spec.data_addr = fpcr & 0x80 ? 0x7C00 : 0x0000;
      device_spec.data_size = fpcr & 0x80 ? 0x0400 : 0x0000;
    }
    else
    {
      rv = 1;
    }
    //
    // save device dsys
    //
    if (rv == 0)
      save_device_dsys(handle);
  }
  return rv;
}

void show_device_details()
{
  printf("Target Device: %s\n\n", device_spec.name);
  printf("  DSYS: %04X - %04X (%04X Bytes)\n", device_spec.dsys_addr, device_spec.dsys_addr + device_spec.dsys_size, device_spec.dsys_size);
  printf("  INFO: %04X - %04X (%04X Bytes)\n", device_spec.info_addr, device_spec.info_addr + device_spec.info_size, device_spec.info_size);
  printf("  CODE: %04X - %04X (%04X Bytes)\n", device_spec.code_addr, device_spec.code_addr + device_spec.code_size, device_spec.code_size);
  if (device_spec.data_size)
    printf("  DATA: %04X - %04X (%04X Bytes)\n", device_spec.data_addr, device_spec.data_addr + device_spec.data_size, device_spec.data_size);
  else
    printf("  DATA: Disabled. (FPCR.DAEN = 0)\n");
  printf("\nPin Connections:\n\n%s\n", device_spec.pins);
}

enum
{
  CMD_EXIT,
  CMD_TEST,
  CMD_HELP,
  CMD_CLEAR,
  CMD_DUMP,
  CMD_DEVICE_HELP,
  CMD_DEVICE_NRF24LE1,
  CMD_DEVICE_NRF24LU1P16,
  CMD_DEVICE_NRF24LU1P32,
  CMD_READ_HELP,
  CMD_READ_DATA,
  CMD_READ_CODE,
  CMD_READ_INFO,
  CMD_READ_DSYS,
  CMD_READ_FSR,
  CMD_READ_FPCR,
  CMD_PROG_HELP,
  CMD_PROG_CODE,
  CMD_PROG_DATA,
  CMD_PROG_INFO,
  CMD_PROG_DSYS,
  CMD_VERIFY_HELP,
  CMD_VERIFY_CODE,
  CMD_VERIFY_DATA,
  CMD_VERIFY_INFO,
  CMD_VERIFY_DSYS,
  CMD_ERASE_HELP,
  CMD_ERASE_CODE,
  CMD_ERASE_DATA,
  CMD_ERASE_INFO,
  CMD_ERASE_INFO_DSYS,
  CMD_ERASE_ALL,
  CMD_RDIS_HELP,
  CMD_RDIS_MB,
  CMD_RDIS_IP,
  CMD_LOAD_HELP,
  CMD_LOAD,
  CMD_SAVE_HELP,
  CMD_SAVE,
  CMD_EDIT_HELP,
  CMD_EDIT,
  CMD_FILL_HELP,
  CMD_FILL,
  CMD_COPY_HELP,
  CMD_COPY,
  CMD_MOVE_HELP,
  CMD_MOVE,
};

const struct
{
  int   id;
  const char *args[5];
}
command_list[] =
{
  { CMD_EXIT              ,{ "EXIT"  , NULL       , NULL  , NULL , NULL } },
  { CMD_EXIT              ,{ "QUIT"  , NULL       , NULL  , NULL , NULL } },
  { CMD_TEST              ,{ "TEST"  , NULL       , NULL  , NULL , NULL } },
  { CMD_HELP              ,{ "HELP"  , NULL       , NULL  , NULL , NULL } },
  { CMD_HELP              ,{ "?"     , NULL       , NULL  , NULL , NULL } },
  { CMD_CLEAR             ,{ "CLEAR" , NULL       , NULL  , NULL , NULL } },
  { CMD_CLEAR             ,{ "@"     , NULL       , NULL  , NULL , NULL } },
  { CMD_DUMP              ,{ "DUMP"  , NULL       , NULL  , NULL , NULL } },
  { CMD_DUMP              ,{ "."     , NULL       , NULL  , NULL , NULL } },
  { CMD_DEVICE_NRF24LE1   ,{ "DEVICE", name_24le1 , NULL  , NULL , NULL } },
  { CMD_DEVICE_NRF24LU1P16,{ "DEVICE", name_lu1p16, NULL  , NULL , NULL } },
  { CMD_DEVICE_NRF24LU1P32,{ "DEVICE", name_24lu1p, NULL  , NULL , NULL } },
  { CMD_DEVICE_HELP       ,{ "DEVICE", "%*"       , NULL  , NULL , NULL } },
  { CMD_READ_CODE         ,{ "READ"  , "CODE"     , NULL  , NULL , NULL } },
  { CMD_READ_DATA         ,{ "READ"  , "DATA"     , NULL  , NULL , NULL } },
  { CMD_READ_INFO         ,{ "READ"  , "INFO"     , NULL  , NULL , NULL } },
  { CMD_READ_DSYS         ,{ "READ"  , "DSYS"     , NULL  , NULL , NULL } },
  { CMD_READ_FSR          ,{ "READ"  , "FSR"      , NULL  , NULL , NULL } },
  { CMD_READ_FPCR         ,{ "READ"  , "FPCR"     , NULL  , NULL , NULL } },
  { CMD_READ_HELP         ,{ "READ"  , "%*"       , NULL  , NULL , NULL } },
  { CMD_PROG_CODE         ,{ "PROG"  , "CODE"     , NULL  , NULL , NULL } },
  { CMD_PROG_DATA         ,{ "PROG"  , "DATA"     , NULL  , NULL , NULL } },
  { CMD_PROG_INFO         ,{ "PROG"  , "INFO"     , NULL  , NULL , NULL } },
  { CMD_PROG_DSYS         ,{ "PROG"  , "DSYS"     , NULL  , NULL , NULL } },
  { CMD_PROG_HELP         ,{ "PROG"  , "%*"       , NULL  , NULL , NULL } },
  { CMD_VERIFY_CODE       ,{ "VERIFY", "CODE"     , NULL  , NULL , NULL } },
  { CMD_VERIFY_DATA       ,{ "VERIFY", "DATA"     , NULL  , NULL , NULL } },
  { CMD_VERIFY_INFO       ,{ "VERIFY", "INFO"     , NULL  , NULL , NULL } },
  { CMD_VERIFY_DSYS       ,{ "VERIFY", "DSYS"     , NULL  , NULL , NULL } },
  { CMD_VERIFY_HELP       ,{ "VERIFY", "%*"       , NULL  , NULL , NULL } },
  { CMD_ERASE_CODE        ,{ "ERASE" , "CODE"     , "ALL" , NULL , NULL } },
  { CMD_ERASE_CODE        ,{ "ERASE" , "CODE"     , NULL  , NULL , NULL } },
  { CMD_ERASE_DATA        ,{ "ERASE" , "DATA"     , "ALL" , NULL , NULL } },
  { CMD_ERASE_DATA        ,{ "ERASE" , "DATA"     , NULL  , NULL , NULL } },
  { CMD_ERASE_INFO_DSYS   ,{ "ERASE" , "INFO"     , "DSYS", NULL , NULL } },
  { CMD_ERASE_INFO        ,{ "ERASE" , "INFO"     , NULL  , NULL , NULL } },
  { CMD_ERASE_ALL         ,{ "ERASE" , "ALL"      , NULL  , NULL , NULL } },
  { CMD_ERASE_HELP        ,{ "ERASE" , "%*"       , NULL  , NULL , NULL } },
  { CMD_RDIS_MB           ,{ "RDIS"  , "MB"       , NULL  , NULL , NULL } },
  { CMD_RDIS_IP           ,{ "RDIS"  , "IP"       , NULL  , NULL , NULL } },
  { CMD_RDIS_HELP         ,{ "RDIS"  , "%*"       , NULL  , NULL , NULL } },
  { CMD_LOAD              ,{ "LOAD"  , "%s"       , NULL  , NULL , NULL } },
  { CMD_LOAD_HELP         ,{ "LOAD"  , "%*"       , NULL  , NULL , NULL } },
  { CMD_SAVE              ,{ "SAVE"  , "%s"       , NULL  , NULL , NULL } },
  { CMD_SAVE_HELP         ,{ "SAVE"  , "%*"       , NULL  , NULL , NULL } },
  { CMD_EDIT              ,{ "EDIT"  , "%4x"      , "%2x*", NULL , NULL } },
  { CMD_EDIT              ,{ "EDIT"  , "%4x"      , NULL  , NULL , NULL } },
  { CMD_EDIT_HELP         ,{ "EDIT"  , "%*"       , NULL  , NULL , NULL } },
  { CMD_FILL              ,{ "FILL"  , "%4x"      , "%2x" , "%4x", NULL } },
  { CMD_FILL_HELP         ,{ "FILL"  , "%*"       , NULL  , NULL , NULL } },
  { CMD_COPY              ,{ "COPY"  , "%4x"      , "%4x" , "%4x", NULL } },
  { CMD_COPY_HELP         ,{ "COPY"  , "%*"       , NULL  , NULL , NULL } },
  { CMD_MOVE              ,{ "MOVE"  , "%4x"      , "%4x" , "%4x", NULL } },
  { CMD_MOVE_HELP         ,{ "MOVE"  , "%*"       , NULL  , NULL , NULL } },
};

int isxdigitstr(char *str, unsigned long width)
{
  if (width == 0)
    width = sizeof(width) << 1;
  do
  {
    if ((width-- == 0) || !isxdigit(*str++))
      return FALSE;
  }
  while (*str);
  return TRUE;
}

int isdigitstr(char *str, unsigned long width)
{
  if (_strnicmp(str, "0x", 2) == 0)
    return isxdigitstr(str + 2, width);
  else
  {
    if (width == 0)
      width = sizeof(width) << 1;
    do
    {
      if ((width-- == 0) || !isdigit(*str++))
        return FALSE;
    }
    while (*str);
  }
  return TRUE;
}

int get_command_id(char **args)
{
  int rv = -1;
  for (int i = 0; (rv < 0) && (i < LENGTHOF(command_list)); ++i)
  {
    rv = command_list[i].id;
    int count = 1;
    for (int j = 0; j < LENGTHOF(command_list[0].args); ++j)
    {
      if (!command_list[i].args[j])
        break;
      count++;
    }
    for (int j = 0; j < count; )
    {
      const char *p = command_list[i].args[j];
      char *q = args[j++];
      if (!p)
      {
        if (q)
        {
          rv = -1;
          break;
        }
      }
      else if (strcmp(p, "%*") == 0)
      {
        break;
      }
      else if (!q)
      {
        rv = -1;
        break;
      }
      else if (p[0] == '%')
      {
        char *e;
        unsigned long w = strtoul(p + 1, &e, 10);
        char c = *e++;
        char r = *e;
        int k = j;
        do
        {
          if (c == 'x')
          {
            if (!isxdigitstr(q, w))
              rv = -1;
          }
          else if (c == 'd')
          {
            if (!isdigitstr(q, w))
              rv = -1;
          }
          else if (c != 's')
            rv = -1;
        }
        while ((r == '*') && (rv >= 0) && (q = args[k++]));
        if (r == '*')
          break;
      }
      else if (stricmp(p, q) != 0)
      {
        rv = -1;
        break;
      }
    }
  }
  return rv;
}

void shell(char **argv)
{
  char cmd[MAX_COMMANDLINE_LENGTH];
  int  len;

#ifdef _MSC_VER
  len = sprintf(cmd, "cmd /C");
#else
  len = sprintf(command, "sh -c");
#endif
  while (*argv)
  {
    char *p = strchr(*argv, ' ') ? " \"%s\"" : " %s";
    if (len + (strlen(p) - 2) + strlen(*argv) >= LENGTHOF(cmd))
      break;
    len += sprintf(cmd + len, p, *argv++);
  }
  system(cmd);
}

char *get_file_ext(char *filename)
{
  char *p = strchr(filename, '.');
  if (!p)
    p = filename + strlen(filename);
  return p;
}

int strtoi(char *str)
{
  return strtoul(str, NULL, 16);
}

int split(char *str, const char *delimters, char **argv, int size)
{
  int count = 0;
  while ((count < (size - 1)) && *(str += strspn(str, delimters)))
  {
    if ((*str == '"') || (*str == '\''))
    {
      char quort[] = {*str++, 0};
      argv[count++] = str;
      str += strcspn(str, quort);
    }
    else
    {
      argv[count++] = str;
      str += strcspn(str, delimters);
    }
    *str++ = 0;
  }
  argv[count] = NULL;
  return count;
}

const char *help_read       = "READ {CODE|DATA|INFO|DSYS|FSR|FPCR}";
const char *help_prog       = "PROG {CODE|DATA|INFO|DSYS}";
const char *help_verify     = "VERIFY {CODE|DATA|INFO|DSYS}";
const char *help_erase_page = "ERASE {CODE|DATA} [ALL]";
const char *help_erase_info = "ERASE INFO [DSYS]";
const char *help_erase_all  = "ERASE ALL";
const char *help_rdis       = "RDIS {MB|IP}";
const char *help_edit       = "EDIT address [value]...";
const char *help_fill       = "FILL destination value count";
const char *help_copy       = "COPY destination source count";
const char *help_move       = "MOVE destination source count";
const char *help_load       = "LOAD filename (.hex or .bin)";
const char *help_save       = "SAVE filename (.hex)";

void help()
{
  printf("Program Commmand List:\n");
  printf("\n");
  printf("  DEVICE [24LE1|LU1P16|24LU1P]\n");
  printf("  %s\n", help_read);
  printf("  %s\n", help_prog);
  printf("  %s\n", help_verify);
  printf("  %s\n", help_erase_page);
  printf("  %s\n", help_erase_info);
  printf("  %s\n", help_erase_all);
  printf("  %s\n", help_rdis);
  printf("  %s\n", help_edit);
  printf("  %s\n", help_fill);
  printf("  %s\n", help_copy);
  printf("  %s\n", help_move);
  printf("  %s\n", help_load);
  printf("  %s\n", help_save);
  printf("  CLEAR\n");
  printf("  DUMP\n");
  printf("  HELP\n");
  printf("  EXIT\n");
  printf("\n");
  printf("  @filename ... execute response file\n");
  printf("\n");
}

int program(FT_HANDLE handle, FILE *fp)
{
  int retcode = 0;
  //
  // Main Loop
  //
  while (1)
  {
    char *argv[MAX_COMMANDLINE_ARGC], command[MAX_COMMANDLINE_LENGTH];
    //
    // command line input
    //
    memset(command, 0, sizeof(command));
    _getcwd(command, sizeof(command) / sizeof(command[0]) - 1);
    printf("%s)", command);
    if (!fgets(command, sizeof(command) - 1, fp))
      strcpy(command, "exit\n");
    //
    // replace command line variables
    //
    replace_vars(command, sizeof(command));
    //
    // print coomand line
    //
    if (fp != stdin)
      printf(command);
    //
    // parse command line
    //
    if (split(command, " \t\r\n", (char **)&argv, sizeof(argv) / sizeof(argv[0])))
    {
      FT_STATUS status;
      uint8 value, *buf;
      int addr, sour, size, id = get_command_id(argv);
      //
      // EXIT
      //
      if (id == CMD_EXIT)
        break;
      //
      // HELP
      //
      else if (id == CMD_HELP)
      {
        help();
        continue;
      }
      //
      // Device Active ?
      //
      else if (!FT_SUCCESS(status = read_fsr(handle, &value)) || (value & FSR_ZERO))
      {
        printf("Device is Not Working.\n");
        continue;
      }
      switch (id)
      {
      //
      // TEST
      //
      case CMD_TEST:
        break;
      //
      // CLEAR
      //
      case CMD_CLEAR:
        get_device_memory(TRUE);
        break;
      //
      // DUMP
      //
      case CMD_DUMP:
        dump_device_memory(16);
        break;
      //
      // DEVICE
      //
      case CMD_DEVICE_HELP:
      case CMD_DEVICE_NRF24LE1:
      case CMD_DEVICE_NRF24LU1P16:
      case CMD_DEVICE_NRF24LU1P32:
        change_device(handle, argv[1]);
        show_device_details();
        break;
      //
      // READ
      //
      case CMD_READ_HELP:
        printf("%s\n", help_read);
        break;
      case CMD_READ_CODE:
        if (!FT_SUCCESS(status = read_flush(handle, device_spec.code_addr, get_device_memory(TRUE) + device_spec.code_addr, device_spec.code_size, 0)))
          printf("Failed. [%d]\n", status);
        else
          dump_device_memory(16);
        break;
      case CMD_READ_DATA:
        if (!FT_SUCCESS(status = read_flush(handle, device_spec.data_addr, get_device_memory(TRUE) + device_spec.data_addr, device_spec.data_size, 0)))
          printf("Failed. [%d]\n", status);
        else
          dump_device_memory(16);
        break;
      case CMD_READ_INFO:
        if (!FT_SUCCESS(status = read_flush(handle, device_spec.info_addr, get_device_memory(TRUE) + device_spec.info_addr, device_spec.info_size, OPT_INFOPAGE)))
          printf("Failed. [%d]\n", status);
        else
          dump_device_memory(16);
        break;
      case CMD_READ_DSYS:
        if (!FT_SUCCESS(status = read_flush(handle, device_spec.dsys_addr, get_device_memory(TRUE) + device_spec.dsys_addr, device_spec.dsys_size, OPT_INFOPAGE)))
          printf("Failed. [%d]\n", status);
        else
          dump_device_memory(16);
        break;
      case CMD_READ_FSR:
        if (!FT_SUCCESS(status = read_fsr(handle, &value)))
          printf("Failed. [%d]\n", status);
        else
          printf("%02X\n", value);
        break;
      case CMD_READ_FPCR:
        if (!FT_SUCCESS(status = read_fpcr(handle, &value)))
          printf("Failed. [%d]\n", status);
        else
          printf("%02X\n", value);
        break;
      //
      // PROG
      //
      case CMD_PROG_HELP:
        printf("%s\n", help_prog);
        break;
      case CMD_PROG_CODE:
        status = FT_OK;
        buf = get_device_memory(FALSE);
        for (addr = device_spec.code_addr; FT_SUCCESS(status) && (addr < device_spec.code_addr + device_spec.code_size); addr += device_spec.page_size)
        {
          for (int i = 0; i < device_spec.page_size; ++i)
          {
            if (buf[addr + i] != 0xFF)
            {
              status = write_flush(handle, addr, buf + addr, device_spec.page_size, 0);
              break;
            }
          }
        }
        if (!FT_SUCCESS(status))
          printf("Failed. [%d]\n", status);
        else
          printf("Ok.\n");
        break;
      case CMD_PROG_DATA:
        status = FT_OK;
        buf = get_device_memory(FALSE);
        for (addr = device_spec.data_addr; FT_SUCCESS(status) && (addr < device_spec.data_addr + device_spec.data_size); addr += device_spec.page_size)
        {
          for (int i = 0; i < device_spec.page_size; ++i)
          {
            if (buf[addr + i] != 0xFF)
            {
              status = write_flush(handle, addr, buf + addr, device_spec.page_size, 0);
              break;
            }
          }
        }
        if (!FT_SUCCESS(status))
          printf("Failed. [%d]\n", status);
        else
          printf("Ok.\n");
        break;
      case CMD_PROG_INFO:
        if (!FT_SUCCESS(status = write_flush(handle, device_spec.info_addr, get_device_memory(FALSE) + device_spec.info_addr, device_spec.info_size, OPT_INFOPAGE)))
          printf("Failed. [%d]\n", status);
        else
          printf("Ok.\n");
        break;
      case CMD_PROG_DSYS:
        if (!FT_SUCCESS(status = write_flush(handle, device_spec.dsys_addr, get_device_memory(FALSE) + device_spec.dsys_addr, device_spec.dsys_size, OPT_INFOPAGE)))
          printf("Failed. [%d]\n", status);
        else
          printf("Ok.\n");
        break;
      //
      // VERIFY
      //
      case CMD_VERIFY_HELP:
        printf("%s\n", help_verify);
        break;
      case CMD_VERIFY_CODE:
        if (!FT_SUCCESS(status = read_flush(handle, device_spec.code_addr, get_verify_memory(TRUE) + device_spec.code_addr, device_spec.code_size, 0)))
          printf("Failed. [%d]\n", status);
        else
        {
          retcode = 0;
          buf = get_device_memory(FALSE);
          for (addr = device_spec.code_addr; addr < device_spec.code_addr + device_spec.code_size; addr += device_spec.page_size)
          {
            for (int i = 0; i < device_spec.page_size; ++i)
            {
              if (buf[addr + i] != 0xFF)
              {
                retcode += verify_device_memory(addr, device_spec.page_size);
                break;
              }
            }
          }
          if (retcode)
            printf("%d Bytes Unmatched.\n", retcode);
          else
            printf("Ok.\n");
        }
        break;
      case CMD_VERIFY_DATA:
        if (!FT_SUCCESS(status = read_flush(handle, device_spec.data_addr, get_verify_memory(TRUE) + device_spec.data_addr, device_spec.data_size, 0)))
          printf("Failed. [%d]\n", status);
        else
        {
          retcode = 0;
          buf = get_device_memory(FALSE);
          for (addr = device_spec.data_addr; addr < device_spec.data_addr + device_spec.data_size; addr += device_spec.page_size)
          {
            for (int i = 0; i < device_spec.page_size; ++i)
            {
              if (buf[addr + i] != 0xFF)
              {
                retcode += verify_device_memory(addr, device_spec.page_size);
                break;
              }
            }
          }
          if (retcode)
            printf("%d Bytes Unmatched.\n", retcode);
          else
            printf("Ok.\n");
        }
        break;
      case CMD_VERIFY_INFO:
        if (!FT_SUCCESS(status = read_flush(handle, device_spec.info_addr, get_verify_memory(TRUE) + device_spec.info_addr, device_spec.info_size, OPT_INFOPAGE)))
          printf("Failed. [%d]\n", status);
        else if (retcode = verify_device_memory(device_spec.info_addr, device_spec.info_size))
          printf("%d Bytes Unmatched.\n", retcode);
        else
          printf("Ok.\n");
        break;
      case CMD_VERIFY_DSYS:
        if (!FT_SUCCESS(status = read_flush(handle, device_spec.dsys_addr, get_verify_memory(TRUE) + device_spec.dsys_addr, device_spec.dsys_size, OPT_INFOPAGE)))
          printf("Failed. [%d]\n", status);
        else if (retcode = verify_device_memory(device_spec.dsys_addr, device_spec.dsys_size))
          printf("%d Bytes Unmatched.\n", retcode);
        else
          printf("Ok.\n");
        break;
      //
      // ERASE
      //
      case CMD_ERASE_HELP:
        printf("%s\n%s\n%s\n", help_erase_page, help_erase_info, help_erase_all);
        break;
      case CMD_ERASE_CODE:
        status = FT_OK;
        buf = get_device_memory(FALSE);
        for (addr = device_spec.code_addr; FT_SUCCESS(status) && (addr < device_spec.code_addr + device_spec.code_size); addr += device_spec.page_size)
        {
          for (int i = 0; i < device_spec.page_size; ++i)
          {
            if (argv[2] || (buf[addr + i] != 0xFF))
            {
              int pageno = addr / device_spec.page_size;
              printf("Erase Page %d\n", pageno);
              status = erase_page(handle, pageno, 0, NULL);
              break;
            }
          }
        }
        if (!FT_SUCCESS(status))
          printf("Failed. [%d]\n", status);
        else
          printf("Ok.\n");
        break;
      case CMD_ERASE_DATA:
        status = FT_OK;
        buf = get_device_memory(FALSE);
        for (addr = device_spec.data_addr; FT_SUCCESS(status) && (addr < device_spec.data_addr + device_spec.data_size); addr += device_spec.page_size)
        {
          for (int i = 0; i < device_spec.page_size; ++i)
          {
            if (argv[2] || (buf[addr + i] != 0xFF))
            {
              int pageno = addr / device_spec.page_size;
              printf("Erase Page %d\n", pageno);
              status = erase_page(handle, pageno, 0, NULL);
              break;
            }
          }
        }
        if (!FT_SUCCESS(status))
          printf("Failed. [%d]\n", status);
        else
          printf("Ok.\n");
        break;
      case CMD_ERASE_INFO:
        if (FT_SUCCESS(status = erase_page(handle, device_spec.info_addr / device_spec.page_size, OPT_INFOPAGE, device_spec.name)))
          printf("Ok.\n");
        else if (status == FT_OTHER_ERROR)
          printf("ERASE ALL command must be executed.\n");
        else
          printf("Failed. [%d]\n", status);
        break;
      case CMD_ERASE_INFO_DSYS:
        if (FT_SUCCESS(status = erase_page(handle, device_spec.dsys_addr / device_spec.page_size, OPT_INFOPAGE | OPT_DSYS, device_spec.name)))
          printf("Ok.\n");
        else if (status == FT_OTHER_ERROR)
          printf("ERASE ALL command must be executed.\n");
        else
          printf("Failed. [%d]\n", status);
        break;
      case CMD_ERASE_ALL:
        if (!FT_SUCCESS(status = erase_all(handle)))
          printf("Failed. [%d]\n", status);
        else
          printf("Ok.\n");
        break;
      //
      // RDIS
      //
      case CMD_RDIS_HELP:
        printf("%s\n", help_rdis);
        break;
      case CMD_RDIS_MB:
        if (!FT_SUCCESS(status = readback_disable(handle, 0)))
          printf("Failed. [%d]\n", status);
        else
          printf("Ok.\n");
        break;
      case CMD_RDIS_IP:
        if (!FT_SUCCESS(status = readback_disable(handle, OPT_INFOPAGE)))
          printf("Failed. [%d]\n", status);
        else
          printf("Ok.\n");
        break;
      //
      // EDIT
      //
      case CMD_EDIT_HELP:
        printf("%s\n", help_edit);
        break;
      case CMD_EDIT:
        addr = strtoi(argv[1]);
        buf = get_device_memory(FALSE) + addr;
        printf("%04X:", addr);
        if (argv[2])
        {
          for (int i = 2; argv[i]; )
          {
            uint8 b = strtoi(argv[i++]);
            *buf++ = b;
            printf(" %02X", b);
          }
          printf("\n");
        }
        else
        {
          printf(" %02X\n", *buf);
        }
        break;
      //
      // FILL
      //
      case CMD_FILL_HELP:
        printf("%s\n", help_fill);
        break;
      case CMD_FILL:
        addr = strtoi(argv[1]);
        value = strtoi(argv[2]);
        size = strtoi(argv[3]);
        buf = get_device_memory(FALSE);
        memset(buf + addr, value, size);
        break;
      //
      // COPY
      //
      case CMD_COPY_HELP:
        printf("%s\n", help_copy);
        break;
      case CMD_COPY:
        addr = strtoi(argv[1]);
        sour = strtoi(argv[2]);
        size = strtoi(argv[3]);
        buf = get_device_memory(FALSE);
        memmove(buf + addr, buf + sour, size);
        break;
      //
      // MOVE
      //
      case CMD_MOVE_HELP:
        printf("%s\n", help_move);
        break;
      case CMD_MOVE:
        addr = strtoi(argv[1]);
        sour = strtoi(argv[2]);
        size = strtoi(argv[3]);
        buf = get_device_memory(FALSE);
        memmove(buf + addr, buf + sour, size);
        if (addr <= sour)
        {
          if (sour - addr < size)
            memset(buf + addr + size, 0xFF, sour - addr);
          else
            memset(buf + sour, 0xFF, size);
        }
        else
        {
          if (addr - sour < size)
            memset(buf + sour, 0xFF, addr - sour);
          else
            memset(buf + sour, 0xFF, size);
        }
        break;
      //
      // LOAD
      //
      case CMD_LOAD_HELP:
        printf("%s\n", help_load);
        break;
      case CMD_LOAD:
        if (_stricmp(get_file_ext(argv[1]), ".bin") == 0 ? load_bin_file(argv[1]) : load_hex_file(argv[1]))
          printf("Failed.\n");
        else
          printf("Ok.\n");
        break;
      //
      // SAVE
      //
      case CMD_SAVE_HELP:
        printf("%s\n", help_save);
        break;
      case CMD_SAVE:
        if (save_hex_file(argv[1], 16))
          printf("Failed.\n");
        else
          printf("Ok.\n");
        break;
      //
      // etc.
      //
      default:
        //
        // execute response file
        //
        if (*argv[0] == '@')
        {
          FILE *fp;
          if (fp = fopen(argv[0] + 1, "rt"))
          {
            retcode = program(handle, fp);
            fclose(fp);
          }
          else
            printf("Response File Can't Open [%s]\n", argv[0] + 1);
        }
        //
        // change directory
        //
        else if (_stricmp(argv[0], "CD") == 0)
        {
          if (!argv[1])
            shell(argv);
          else
            _chdir(argv[1]);
        }
#ifdef _MSC_VER
        //
        // change drive (windows only)
        //
        else if ((strlen(argv[0]) == 2) && isalpha(argv[0][0]) && (argv[0][1] == ':'))
          _chdir(argv[0]);
#endif
        //
        // execute shell
        //
        else
          shell(argv);
        break;
      }
    }
  }
  return retcode;
}

int select_usb_device(char *description, char *serialnumber)
{
  uint32 channels;
  int index = 0;
  //
  // Get USB Device Count
  //
  if (FT_SUCCESS(get_usb_device_count(&channels)))
  {
    if (!description)
      printf("Number of USB Devices: %d\n", (int)channels);
    if (channels > 0)
    {
      //
      // Display USB Device Info
      //
      for (uint32 i = 0; i < channels; i++)
      {
        FT_DEVICE_LIST_INFO_NODE devList;
        if (FT_SUCCESS(get_usb_device_info(i, &devList)))
        {
          if (!description)
            printf("[%d] \"%s\" \"%s\" (%04X)\n", i + 1, devList.Description, devList.SerialNumber, devList.LocId);
          if (description && (stricmp(description, devList.Description) == 0))
          {
            if (!serialnumber || (stricmp(serialnumber, devList.SerialNumber) == 0))
              return i + 1;
          }
        }
      }
      //
      // Select USB Device
      //
      if (!description)
      {
        char c;
        printf("Select USB Device Number [0:Quit, 1 - %X] ? ", channels);
        do
        {
          index = -1;
          c = _getch();
          if (isdigit(c))
            index = c - '0';
          else if (isxdigit(c))
            index = _toupper(c) - 'A' + 10;
        } while ((index < 0) || (index > (int)channels));
        printf("%c\n\n", c);
      }
    }
  }
  return index;
}

int main(int argc, char **argv)
{
  const char *targetdevice = name_24le1;
  char *responsefile = NULL;
  char *description  = NULL;
  char *serialnumber = NULL;
  int usage, index, retcode = 0;
  FILE *response = NULL;

  printf("Nordic nRF24LE1/LU1P SPI Programmer Shell for FTDI MPSSE.\n");
  printf("Version 1.0, (C) 2017 Sasapea's Lab.\n\n");

  //
  // Init Variables
  //
  init_vars();
  //
  // Command Line Options
  //
  usage = 0;
  for (int i = 1; i < argc; )
  {
    char *p = argv[i++];
    if (*p == '-')
    {
      if (i < argc)
      {
        ++p;
        if (stricmp(p, "d") == 0)
          description = argv[i++];
        else if (stricmp(p, "s") == 0)
          serialnumber = argv[i++];
        else if (stricmp(p, "t") == 0)
          targetdevice = argv[i++];
        else
          usage = 1;
      }
      else
        usage = 1;
    }
    else if (*p == '@')
    {
      if (!responsefile)
        responsefile = p + 1;
      else
        usage = 1;
    }
    else
    {
      //
      // Add Variables
      //
      add_vars(p);
    }
  }
  //
  // Check Environment Variables
  //
  if (getenv(ENV_INFO_PATH) == NULL)
    printf("Please set the information save destination path to the environment variable \"%s\".\n\n", ENV_INFO_PATH);

  //
  // Usage and Exit
  //
  if (usage)
  {
    printf("Usage: nrf24prog [Option...] [@ResponseFile]\n");
    printf("\n");
    printf("  Option:\n");
    printf("    -d Description  ... Description  of USB Device\n");
    printf("    -s SerialNumber ... SerialNumber of USB Device\n");
    printf("    -t TargetDevice ... Target Device {%s|%s|%s}\n", name_24le1, name_lu1p16, name_24lu1p);
    printf("\n");
    free_vars();
    return 1;
  }

  //
  // Open Response File
  //
  if (responsefile)
  {
    if ((response = fopen(responsefile, "rt")) == NULL)
    {
      printf("Response File Can't Open [%s]\n", responsefile);
      free_vars();
      return 1;
    }
  }

  //
  // Init MPSSE
  //
  Init_libMPSSE();

  //
  // Select USB Device
  //
  index = select_usb_device(description, serialnumber);
  if (index)
  {
    //
    // Open USB Device
    //
    FT_HANDLE handle;
    if (FT_SUCCESS(open_usb_device(&handle, index - 1)))
    {
      //
      // Init SPI Mode
      //
      if (FT_SUCCESS(init_spi(handle)))
      {
        //
        // Program Mode Start (PROG = HIGH)
        //
        if (FT_SUCCESS(program_mode(handle, TRUE)))
        {
          char device[16];
          //
          // Check Device Info
          //
          if (check_device_dsys(handle, device))
            strcpy(device, targetdevice);
          //
          // Set Target Device
          //
          if (change_device(handle, device) != 0)
            change_device(handle, name_24le1);
          show_device_details();
          //
          // Init Device Memory
          //
          get_device_memory(TRUE);
          //
          // HELP
          //
          if (!response)
            help();
          //
          // Program
          //
          retcode = program(handle, response ? response : stdin);
          //
          // Program Mode Exit
          //
          program_mode(handle, FALSE);
        }
      }
      //
      // Close USB Device
      //
      close_usb_device(handle);
    }
  }

  //
  // Cleanup MPSSE
  //
  Cleanup_libMPSSE();

  //
  // Close Response File
  //
  if (response)
    fclose(response);

  //
  // Free Variables
  //
  free_vars();

  return retcode;
}

