#ifndef __NRF24SPI_H
#define __NRF24XX_H

#include "libMPSSE_spi.h"

//
// SPI init parameters
//
#define SPI_CLOCK_RATE    2000000  // 2MHz
#define SPI_LATENCY_TIMER 2
//
// SPI Pin Assign (C232HM)
//
#define SPI_SCK     0  // Orange
#define SPI_MOSI    1  // Yellow
#define SPI_MISO    2  // Green
#define SPI_CSN     3  // Brown
#define SPI_PROG    6  // White
#define SPI_RESET   7  // Blue
//
// Pin mode
//
#define INPUT  0
#define OUTPUT 1
//
// GPIO value
//
#define LOW    0
#define HIGH   1
//
// read_flush/write_flush options
//
#define OPT_INFOPAGE 1
#define OPT_DSYS     2
//
// FSR Register Bits
//
#define FSR_DEBUG  0x80
#define FSR_STP    0x40
#define FSR_WEN    0x20
#define FSR_RDYN   0x10
#define FSR_INFEN  0x08
#define FSR_RDISMB 0x04
#define FSR_RDISIP 0x02
#define FSR_ZERO   0x01
//
// DSYS Size
//
#define DSYS_SIZE  0x20
//
// Functions
//
FT_STATUS pin_mode(FT_HANDLE handle, uint8 pin, uint8 mode);
FT_STATUS pin_write(FT_HANDLE handle, uint8 pin, uint8 value);
FT_STATUS pin_read(FT_HANDLE handle, uint8 pin, uint8 *value);
FT_STATUS reset(FT_HANDLE handle, bool state);
FT_STATUS prog(FT_HANDLE handle, bool state);
FT_STATUS program_mode(FT_HANDLE handle, bool enable);
FT_STATUS read_fsr(FT_HANDLE handle, uint8 *value);
FT_STATUS read_fpcr(FT_HANDLE handle, uint8 *value);
FT_STATUS read_flush(FT_HANDLE handle, uint16 addr, uint8 *buffer, uint16 buflen, uint8 options);
FT_STATUS write_flush(FT_HANDLE handle, uint16 addr, uint8 *buffer, uint16 buflen, uint8 options);
FT_STATUS enable_debug(FT_HANDLE handle);
FT_STATUS readback_disable(FT_HANDLE handle, uint8 options);
FT_STATUS erase_all(FT_HANDLE handle);
FT_STATUS erase_page(FT_HANDLE handle, uint8 pageno, uint8 options, const char *device);
FT_STATUS get_usb_device_count(uint32 *count);
FT_STATUS get_usb_device_info(uint32 index, FT_DEVICE_LIST_INFO_NODE *info);
FT_STATUS open_usb_device(FT_HANDLE *handle, uint32 index);
FT_STATUS close_usb_device(FT_HANDLE handle);
FT_STATUS init_spi(FT_HANDLE handle);

#endif // __NRF24XX_H
