#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

static char  *vars_buffer = NULL;
static size_t vars_length = 0;

#ifdef _MSC_VER
#pragma warning(disable : 4996)
#endif

void free_vars()
{
  if (vars_buffer)
  {
    free(vars_buffer);
    vars_buffer = NULL;
    vars_length = 0;
  }
}

void init_vars()
{
  free_vars();
  vars_buffer = (char *)malloc(32);
  vars_buffer[0] = vars_buffer[1] = 0;
}

int add_vars(const char *str)
{
  size_t len = strlen(str);
  if (len > 0)
  {
    if (len + vars_length + 2 >= (vars_buffer ? _msize(vars_buffer) : 0))
    {
      char *newbuf = (char *)realloc(vars_buffer, vars_length + __max(len, 16) * 2);
      if (newbuf == NULL)
        return 1; // error
      vars_buffer = newbuf;
    }
    strcpy(&vars_buffer[vars_length], str);
    vars_length += len + 1;
    vars_buffer[vars_length] = 0;
  }
  return 0;
}

char *get_vars(const char *name)
{
  char *p = vars_buffer;
  if (p)
  {
    while (*p)
    {
      size_t name_len = strlen(name);
      size_t vars_len = strcspn(p, "=");
      if ((vars_len == name_len) && (_strnicmp(p, name, name_len) == 0))
        return p + vars_len + (p[vars_len] ? 1 : 0);
      p += strlen(p) + 1;
    }
  }
  return getenv(name);
}

int replace_vars(char *str, size_t size)
{
  for (char *q, *p = str; *p; ++p)
  {
    if ((p = strchr(p, '$')) == NULL)
      break;
    if (*(p + 1) == '{')
    {
      if ((q = strchr(p + 2, '}')) != NULL)
      {
        size_t value_len, name_len = (size_t)q - (size_t)p + 1;
        char *value, *name = (char *)_alloca(name_len);
        sprintf(name, "%.*s", (int)(name_len - 3), p + 2);
        if ((value = get_vars(name)) == NULL)
          value = "";
        value_len = strlen(value);
        if (name_len > value_len)
          strcpy(p, p + (name_len - value_len));
        else if (value_len > name_len)
        {
          if (strlen(str) + (value_len - name_len) + 1 > size)
            return 1; // error
          memmove(p + (value_len - name_len), p, strlen(p) + 1);
        }
        memcpy(p, value, value_len);
        p = q;
      }
    }
  }
  return 0;
}
